/*
    dp : An implementation of the davis puttnam procedure for solving
         propositional satisfiability problems.
    Copyright (C) 1993  Peter Barth

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/
/* ------------------------------------------------------------------
  datadefs.h
* -------------------------------------------------------------------
  Global Data structure definition and other global stuff
* ---------------------------------------------------------------- */

#include "datadefs.h"

/******************************************************/
/* The datastructures                                 */
/******************************************************/

list problem = NULL;


/* a) static datastructure */
list variable_references = NULL;
int** lit_to_clauses = NULL; /* The variable references as array for faster acces */
int** clause_to_lits = NULL; /* the clause_id references */

/* b) dynamic datastructure  */
list active_clauses = NULL;     /* int list */
list active_variables = NULL;   /* int list */
silist av,ac;

int* ucvs;   
int* ucvsptr;
int* ucvscmpptr;

list solution = NULL;
list solution_all = NULL;

/* Some global Flags and statistic information */
int showsolution = 0;
int verbosity = 1;
int timing    = 0;
int heuristic = 0;
int no_nodes = 0;
int no_ures = 0;



/* misc. */

void error(char *s)
{
  fprintf(stderr,"\n%s : aborted",s);
  exit(-1);
}

/* converts a double linked list to a single linked list assuming that the
 * data is an integer.
 */
silist list2silist(list l)
{
  silist si=init_silist();
  silist si_act = si;

  l = l_first(l);
  while(l != NULL)
    {
      si_act = si_insafter(si_act,(int) l->data);
      l = l_next(l);
    }
  return si;
}


/* timing */
time_t time(time_t*);

int time_init_ok = 0;
struct tms ref_time;
time_t for_rand_init;

void init_time()
{
  time_init_ok = 1;
  times(&ref_time);
}

int get_time()
{
  struct tms t;
  int time;

  if (time_init_ok == 0)
    init_time();
  times(&t);
  time = (int) ( ((long) t.tms_utime) - ((long) ref_time.tms_utime));
#ifdef linux
  return(time * 10); /* HZ is 100 */
#else
  return(time * 1000 / 60); /* normally HZ is 60 */
#endif
}


/* data structure initialization and destruction */

int uvr_equal(void* d1,void* d2)
{
  return ( ((tvref *)d1)->variable == ((tvref *)d2)->variable );
}

void update_variable_references(int lit,int clause)
{
  tvref svref;
  list l;

  svref.variable = abs(lit);
  svref.clauses = NULL;
  l = l_search(l_first(variable_references),(void *) &svref,uvr_equal);
  
  if (l == NULL) /* variable not yet in list */
  {
    tvref* pvref = malloc(sizeof(tvref));
    
    pvref->variable = abs(lit);
    pvref->clauses = NULL;
    pvref->clauses = l_dinsa(pvref->clauses,
                            (void *) (lit > 0 ? clause : -clause));
    variable_references = l_dinsa(l_last(variable_references),pvref);
    /* initalise active_variables */
    active_variables = l_dinsa(active_variables,(void *) abs(lit));
  }
  else
  {
    tvref* pvref = (tvref *) l_rdata(l);
    pvref->clauses = l_dinsa(pvref->clauses,    
                             (void *) (lit > 0 ? clause : -clause));
  }  
}

void init_lit_to_clauses()
{
  list vr = l_first(variable_references);
  int  max_index = 0;
  int i;

  while(vr)
  {
    tvref *pvref = l_rdata(vr);

    max_index = max(max_index,pvref->variable);
    vr = l_next(vr);
  }
  vr = l_first(variable_references);
  lit_to_clauses = malloc(sizeof(int *) * (max_index + 2));
  if (lit_to_clauses == NULL)
    error("out of memory");
  for (i=1;i <=max_index+1;i++)
    lit_to_clauses[i] = NULL;
  while (vr)
  {
    list clauses = l_first(((tvref *)l_rdata(vr))->clauses);
    int  index   = ((tvref *)l_rdata(vr))->variable;
    int c_len = l_len(clauses);
    int j=0;

    lit_to_clauses[index] = malloc(sizeof(int) * (c_len + 1));
    if (lit_to_clauses[index] == NULL)
      error("out of memory");
    lit_to_clauses[index][c_len] = 0;
    while (clauses)
    {
      lit_to_clauses[index][j++] = (int) l_rdata(clauses);
      clauses = l_next(clauses);
    }
    vr = l_next(vr);
  }
}

int intcmp(void *d1,void *d2)
{
  return (abs((int) d1) > abs((int) d2));
}

void set_clause_to_lits(int clause_id,list clause,int clauselen)
{
  int* cclauseptr;
  clause_to_lits[clause_id] = malloc(sizeof(int)*clauselen+1);
  if (clause_to_lits[clause_id] == NULL)
    error("out of memory");
  cclauseptr = clause_to_lits[clause_id];
  while(clause)
    {
      *cclauseptr++ = (int) l_rdata(clause);
      clause = l_next(clause);
    }
  *cclauseptr = 0;
}

void init_datastructures()
{
  list p = l_first(problem);
  int no_clauses = l_len(p);
  int clause_id;
  int paclause = 0; 
  list vr;

  clause_to_lits = malloc(sizeof(int**)*(no_clauses+2));
  clause_to_lits[no_clauses+1] = NULL;

  while(p)
  {
    list clause = l_first((list) l_rdata(p));
    clause_id = (int) l_rdata(clause);
    set_clause_to_lits(clause_id,l_next(clause),l_len(l_next(clause)));
    SETtaclauseid(paclause,clause_id); 
    SETtaclauselen(paclause,l_len(l_next(clause))); 
    active_clauses = l_dinsa(active_clauses,(void *)paclause);
    clause = l_next(clause);
    while(clause)
    {
       update_variable_references((int) l_rdata(clause),clause_id);
       clause = l_next(clause);
    }
    p = l_next(p);  
  }
  vr = l_first(variable_references);
  while(vr)
  {
    tvref *pvref = (tvref *) l_rdata(vr);
    pvref->clauses = l_first(pvref->clauses);
    vr = l_next(vr);
  } 
  variable_references = l_first(variable_references);
  active_clauses = l_first(active_clauses);
  active_variables = l_sort(l_first(active_variables),intcmp);
  ucvsptr = ucvs = malloc(sizeof(int) * (l_len(active_variables) + 1));
  *ucvsptr = 0;
  init_lit_to_clauses(); /* initialise the lit_to_clauses (variable references) */
  av = list2silist(active_variables);
  ac = list2silist(active_clauses);
  /* Initialize the random number generator */
  time(&for_rand_init);
  srand((int) for_rand_init);
}

void clean_datastructures()
{
  int** ptr;

  free_silist(av);
  free_silist(ac);
  free(ucvs);
  ptr = clause_to_lits+1;
  while (*ptr)
    free(*ptr++);
  free(clause_to_lits);
  ptr = lit_to_clauses+1;
  while (*ptr)
    if (*ptr != NULL)
      free(*ptr++);
    else
      ptr++;
  free(lit_to_clauses);
}
