/*
    dp : An implementation of the davis puttnam procedure for solving
         propositional satisfiability problems.
    Copyright (C) 1993  Peter Barth

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/
/* ------------------------------------------------------------------
  dpio.c
* -------------------------------------------------------------------
  Input / Output  routines for dp
* ---------------------------------------------------------------- */

#include "datadefs.h"
#include "dpio.h"
#include <string.h>
#include <ctype.h>

int allow_zero=1;

void skip_chars(FILE* file,int skip)
{
  int c = fgetc(file);
  if (c == skip)
    skip_chars(file,skip);
  else
    ungetc(c,file);
}

void skip_eoln(FILE* file)
{
  skip_chars(file,'\n');
}

int next_char_is(FILE* file,int is)
{
  int c = fgetc(file);

  ungetc(c,file);
  if (c == is)
    return 1;
  else
    return 0;
}

int feoln(FILE* file)
{
  int c = fgetc(file);

  ungetc(c,file);
  if (c == '\n')
    return 1;
  else
    return 0;
}

/* I hate to write parsers, 
 * somebody like to write a lex front ? you're welcome ! */
int get_int(FILE* file)
{
  char buffer[20];
  char*bufptr = buffer;
  int c;
  int var;
  int sign = 0;
  
  do {
    c = fgetc(file);
    if (sign == 0)      /* only for the zero */
      if (c == '-') sign = -1; else sign = 1;
    if (isalpha(c))     /* skip characters */
      c = fgetc(file);
    *bufptr++ = c;
  } while ((c >= '0' && c <= '9') || c == '-');
  ungetc(*(--bufptr),file);
  *bufptr = 0x00;
  var = atoi(buffer);
  if (!allow_zero)
    if (var == 0)
      return 0;
  if (var == 0)
    return sign;
  if (var >= 0)
    return var+1;
  else
    return var-1;
}

int get_variable(FILE* file)
{ 
  int variable;
  int comment = 1;

  do {
    if (next_char_is(file,'*')) /* comment */
      while (fgetc(file) != '\n'); /* forget line */
    else if (next_char_is(file,'c')) /* comment */
      while (fgetc(file) != '\n'); /* forget line */
    else if (next_char_is(file,'p')) /* comment */
      while (fgetc(file) != '\n'); /* forget line */
    else if (next_char_is(file,'%')) /* comment */
      while (fgetc(file) != '\n'); /* forget line */
    else comment = 0;
  } while (comment);
  if (next_char_is(file,'E')) /* maybe ENDATA */
  { 
    char buffer[80];
    
    fgets(buffer,80,file);
    if (!strncmp("ENDATA",buffer,6))
    {
      while (!feof(file)) fgetc(file);
      return 0;
    }
    else
      error("Line must not begin with 'E'");
  }
#if 0 
  if (next_char_is(file,'0')) /* maybe END OF PROBLEM */
  { 
    skip_chars(file,'0');
    skip_chars(file,' ');
    if (next_char_is(file,'\n')) /* it is the end of the problem */
    {
      while (!feof(file)) fgetc(file);
      return 0;
    }
    /* it is the variable 0 */
    return 0+1;  /* increase the variable by 1 */
  }
#endif
  skip_chars(file,' ');
  skip_chars(file,'|');
  skip_chars(file,' ');
  skip_chars(file,'(');   /* forget parentheses */
  skip_chars(file,' ');
  skip_chars(file,')');   /* forget parentheses */
  skip_chars(file,' ');
  if (!allow_zero)
    {
      skip_chars(file,'\n');
      skip_chars(file,' ');
      skip_chars(file,'|');
      skip_chars(file,' ');
      skip_chars(file,'(');   /* forget parentheses */
      skip_chars(file,' ');
      skip_chars(file,')');   /* forget parentheses */
      skip_chars(file,' ');
      variable = get_int(file);
      return variable;
    }
  else
    {
      if (!next_char_is(file,'\n') && (!next_char_is(file,'.')))
	{
	  variable = get_int(file);
	  return variable;
	}
      skip_chars(file,'.');
      skip_chars(file,'\n');
      return(0);
    }
}

int tautology(list clause)
{
  if (clause == NULL)
    return 0;
  while (clause->next != NULL)
    {
      if ((int) clause->next->data == -(int) clause->data)
	{
	  DO_VERBOSE(10,printf("\ntautology deleted");)
          return 1;
	}
      clause = l_next(clause);
    }
  return 0;
}

list remove_duplicates(list clause)
{
  if (clause == NULL)
    return NULL;
  while (clause->next != NULL)
    {
      if (clause->next->data == clause->data)
	{
	  DO_VERBOSE(10,printf("\ndouble literal deleted");)
          clause = l_del(clause); 
	  if (clause->next == NULL)
	    break;
	}
      else
	clause = l_next(clause);
    }
  return l_first(clause);
}

void read_clauses(FILE *file)
{
  int variable;
  int clause_id = 1;
  list clause=NULL;

  while (!feof(file))
  {
    variable = get_variable(file);
    if (variable) /* NOT END OF CLAUSE */
      clause = l_dinsa(clause,(void *) variable);  
                              /* pointer same size as int !!!!! */
    else
    {
      clause = l_first(clause);
      if (l_len(clause) != 0)
      {
        clause = l_sort(clause,intcmp);
	if (!tautology(clause))
	  {
	    clause = remove_duplicates(clause); 
	    clause = l_dinsb(clause,(void *)clause_id++);
	    problem = l_dinsa(problem,clause);
	  }
      }
      clause = NULL;
    }
  }
}

void write_clause(void *data)
{
  list l = (list) data;

  l = l_first(l);
  printf("  %3i(%2i): ",(int) l_rdata(l),l_len(l_next(l)));
  l = l_next(l);
  while(l)
  {
    int lit = (int) l_rdata(l);  
    printf("%i ",lit<0 ? lit+1 : lit-1); /* as we add one in the parser */
    l = l_next(l);
  }
  printf("\n");
}

void report_problem()
{
  if (verbosity <= 3)
    return;
  printf("\n The problem to solve (%i clauses) \n",l_len(l_first(problem)));
  l_forall(l_first(problem),write_clause);
}

void write_a_v(silist av)
{ 
  av = av->next;
  while(av->next != av) 
    { 
      printf("%i ",av->key-1); /* as we add one in the parser */
      av = av->next;
    } 
}

void write_a_c(silist ac)
{
  ac = ac->next;
  while(ac->next != ac) 
  { 
    printf("%i(%i) ",GETtaclauseid(ac->key),
                     GETtaclauselen(ac->key)); 
    ac = ac->next;
  }
}

void report_datastructures()
{
  list av = l_first(active_variables);

  if (verbosity <= 3)
    return;
  printf("\n The static variable references \n");
  while(av)
  {
    int variable = (int) l_rdata(av);
    int* refs = lit_to_clauses[variable];

    printf("\n  %3i: ",variable-1); /* as we add one in the parser */
    while (*refs)
      printf("%i ",*refs++);
    av = l_next(av);
  }
  { 
    silist ac = list2silist(active_clauses);
    silist av = list2silist(active_variables);
    printf("\n The active clauses   \n   ");
    write_a_c(ac);
    free_silist(ac);
    printf("\n The active variables \n   ");
    write_a_v(av);
    free_silist(av);
  }
}
